

// std c/c++ includes
#include <iostream>
#include <fstream>
#include <sstream>
#include <memory>
#include <string>
#include <iomanip>
#include <cstdlib>
#include <unistd.h>
#include <getopt.h>

// grpc includes
#include <grpcpp/grpcpp.h>
#include "turbo.grpc.pb.h"

// yaml includes
#include <yaml-cpp/yaml.h>


using namespace std;
using namespace grpc;
using namespace Turbo;

#define SETUP_CONN(a,b)                                                      \
	const auto opt_val = getOption(upper_opts, "host");                  \
	const auto creds   = getCredentials(upper_opts);                     \
	auto channel       = grpc::CreateChannel(opt_val,creds);             \
	auto backend       = TurboService::NewStub(channel);                 \
                                                                             \
	auto request = a();                                                  \
	auto reply   = b();                                                  \
	ClientContext context; 

// types 

struct Option_t
{
	string name;
	string value;
};



// typedef for a function pointer like main
using MainLikeFuncPtr_t = int(*)(int argc, char *argv[], const vector<Option_t>& upper_opts) ;

struct SubCommand_t
{
	string            name;
	MainLikeFuncPtr_t func;


};

// would be nice to have in a header, but don't put odd macros in headers
#define ALLOF(a) begin(a),end(a)


string getOption(const vector<Option_t>& opts, const string& name)
{

	const auto found_it = find_if(ALLOF(opts), [&](const Option_t& opt){ return opt.name==name; });
	assert(found_it!=end(opts));
	return found_it->value;

}

int32_t getOptionInt(const vector<Option_t>& opts, const string& name)
{
	return atoi(getOption(opts,name).c_str());
}


// helper functions 
void print_error(const Status& status) 
{
	cout << "Server returned error. " << endl;
	cout << "Number:  " << status.error_code()    << endl; 
	cout << "Message: " << status.error_message() << endl;
	cout << "Details: " << status.error_details() << endl;
}

int Usage(int argc, char* argv[], const vector<Option_t>& upper_opts)
{
	cout << getOption(upper_opts, "pgm-name") << ":" << endl;
	cout << "	--help|-h|--usage|-?               Print this screen"                       << endl;
	cout << "	ping                               Verify that the Turbo server is running" << endl;
	cout << "	boost add <name>                   Create a new boost" << endl;
	cout << "	boost list                         List all existing boosts"<< endl;
	cout << "	boost show <id>                    Show details of a boost" << endl;
	cout << "	boost delete idname>               Delete an existing boost" << endl;
	cout << "	version add <boost id> <file>      Create a new version of a boost" << endl;
	cout << "	version list                       List all existing version of a boost"<< endl;
	cout << "	version show <id>                  Show details of a version of a boost" << endl;
	cout << "	version delete <id>                Delete an existing version of a boost" << endl;
	cout << "	fuzz --fuzz-config <yaml> --app-config <yaml> --ver-id <ver id>                	   " << endl;
	cout << "                                          Start a fuzzing session for version id" << endl;
	cout << "	log get prep <ver id>              Retrieve the prep logs for a version id " << endl;
	cout << "	log get fuzz <ver id>              Retrieve the fuzzing logs for a version id" << endl;
	cout << "	log delete prep <ver id>           Delete the prep logs for a version id" << endl;
	cout << "	log delete fuzz <ver id>           Delete the fuzzing logs for a version id" << endl;
	return 0;
}



shared_ptr<ChannelCredentials> getCredentials(const vector<Option_t>& opts)
{
	return grpc::InsecureChannelCredentials();
}

int subcommandDispatch(string cmd_msg, const vector<SubCommand_t> cmds, int argc, char* argv[], vector<Option_t> upper_opts)
{
	if(argc < 2)
	{
		cout << "Error:  must specify " << cmd_msg << " subcommand "<< endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	const auto subcommand_it = find_if(ALLOF(cmds), 
			[argv](const SubCommand_t& cmd) { return cmd.name == argv[1]; });
	if(subcommand_it ==  end(cmds))
	{
		cout << "Error:  Invalid " << cmd_msg << " subcommand:" << argv[1] << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	return subcommand_it->func(argc-1,argv+1, upper_opts);
}

//
// how to actually parse and perform a (sub)command
// 


// check server alive.
int Ping(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	SETUP_CONN(Empty_t,Empty_t);

	// perform the communication
	const auto status = backend->Ping(&context, request, &reply);

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		return 0;
	} 
	else 
	{
		print_error(status);
		return 1;
	}
}

int VersionAdd(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc<3)
	{
		cout << "Error, version add must specify a file name and boost id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>3)
	{
		cout << "Error, junk at end of version add " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	SETUP_CONN(VersionAddRequest_t,VersionAddReply_t);

	request.set_boost_id(atoi(argv[1]));

	ifstream fin(argv[2], ios::binary);
	if(!fin)
	{
		cout << "Cannot open file: " << argv[2] << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	ostringstream ostrm;
	ostrm << fin.rdbuf();
	const auto &send_contents = ostrm.str();

	request.set_version_contents(send_contents);

	const auto status = backend->VersionAdd(&context, request, &reply);

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		cout << "Version ID: " << reply.version_id() << endl;
		return 0;
	} 
	else 
	{
		print_error(status);
		return 1;
	}
}

int VersionList(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, version list must specify boost id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>2)
	{
		cout << "Error, junk at end of version list " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	SETUP_CONN(VersionListRequest_t,VersionInfo_t);

	request.set_boost_id(atoi(argv[1]));

        auto reader = move(unique_ptr<ClientReader<VersionInfo_t> >( backend->VersionList(&context, request)));


        cout << "id\t" 
	     << setw(16) << left << "orig bin sha"
	     << setw(16) << left << "zafl bin sha"
	     << setw(25) << left << "submit time"
	     << setw(25) << left << "processing start time"
	     << setw(25) << left << "processing end time"
	     << "error" 
	     << endl;
        while(reader->Read(&reply))
        {
                cout << dec 
		     << reply.ver_id        ()<< "\t" 
		     << setw(16) << left << reply.orig_bin_sha  ()
		     << setw(16) << left << reply.zafl_bin_sha  ()
		     << setw(25) << left << reply.submit_time   ()
		     << setw(25) << left << reply.start_time    ()
		     << setw(25) << left << reply.end_time      ()
		     << boolalpha << reply.err()           
                     << endl;

        }

        const auto status = reader->Finish();

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		return 0;
	} 
	else 
	{
		print_error(status);
		return 1;
	}
}

int VersionShow(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

int VersionDelete(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}



//
// process the boost subcommand
//
int Version(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"add",    VersionAdd    },
			{"list",   VersionList   },
			{"show",   VersionShow   },
			{"delete", VersionDelete },
		};

	return subcommandDispatch("version", subcommands, argc, argv, upper_opts);
}


//
// add a "boost" to to the server
// 
int BoostAdd(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, boost add must specify a boost name" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>2)
	{
		cout << "Error, junk at end of boost add " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	SETUP_CONN(BoostAddRequest_t,BoostAddReply_t);

	request.set_name(argv[1]);

	const auto status = backend->BoostAdd(&context, request, &reply);

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		cout << "Boost ID: " << reply.boost_id() << endl;
		return 0;
	}
	else 
	{
		print_error(status);
		return 1;
	}
}


// 
// list all boosts
//
int BoostList(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc>1)
	{
		cout << "Error, junk at end of boost list " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	SETUP_CONN(Empty_t,BoostInfo_t);

	auto reader = move(unique_ptr<ClientReader<BoostInfo_t> >( backend->BoostList(&context, request)));

	cout << "BoostID\tCreate Time\t\t\tName" << endl;
	while(reader->Read(&reply))
	{
		cout << dec 
		     << (reply.id())<< "\t" 
		     << (reply.create_time())<< "\t" 
		     << (reply.name()) 
		     << endl;
		       
	}

	const auto status = reader->Finish();

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		return 0;
	}
	else 
	{
		print_error(status);
		return 1;
	}


	return 1;
}

// 
// show a particular boost
//
int BoostShow(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

//
// remove a boost from the system
//
int BoostDelete(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

//
// process the boost subcommand
//
int Boost(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"add",    BoostAdd    },
			{"list",   BoostList   },
			{"show",   BoostShow   },
			{"delete", BoostDelete },
		};

	return subcommandDispatch("boost", subcommands, argc, argv, upper_opts);
}

//
// process the boost subcommand
//
int Fuzz(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, fuzz subcommand requires version id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	// default options
	auto verbose_flag         = (int)false; // declared int for getopts
	auto fuzz_config_filename = string();
	auto app_config_filename  = string();
	auto ver_id               = -1;

	// parse options
	const struct option long_options[] =
	{
		/* These options set a flag. */
		{"verbose"      , no_argument      , &verbose_flag, 1  },
		{"no-verbose"   , no_argument      , &verbose_flag, 0  },

		/* These options don’t set a flag.  We distinguish them by their indices. */
		{"fuzz-config"  , required_argument, nullptr, 'f'},
		{"app-config"   , required_argument, nullptr, 'a'},
		{"ver-id"       , required_argument, nullptr, 'v'},
		{0, 0, 0, 0}
	};
	const auto short_opts = string("f:a:h?");
	while(1)
	{

		/* getopt_long stores the option index here. */
		auto option_index = 0;
		auto c = getopt_long (argc, argv, short_opts.c_str(), long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
			case 0:
			{
				/* If this option set a flag, do nothing else now. */
				if (long_options[option_index].flag != 0)
					break;

				printf ("option %s", long_options[option_index].name);
				if (optarg)
					printf (" with arg %s", optarg);
				printf ("\n");
				abort();
			}

			case 'f':
			{
				fuzz_config_filename=string(optarg);
				break;
			}

			case 'a':
			{
				app_config_filename=string(optarg);
				break;
			}
			case 'v':
			{
				ver_id=atoi(optarg);
				break;
			}

			case '?':
			case 'h':
			{
				Usage(argc,argv, upper_opts);
				return 1;
			}

			default:
			{
				cout << "Unrecognized option to fuzz" << endl;
				Usage(argc,argv, upper_opts);
				return 1;
			}
		}
	}


	//
	// sanity check and load fuzzing config
	// 
	if(fuzz_config_filename == "" )
	{
		cout<< "Must specify a fuzzing config file " << endl;
		Usage(argc,argv, upper_opts);
		return 1;

	}
	auto _fuzz_config_yaml=YAML::Node(); 
	try
	{
		_fuzz_config_yaml=YAML::LoadFile(fuzz_config_filename);
	}
	catch(const exception& e)
	{
		cout << "Could not parse " << fuzz_config_filename << endl;
		cout << e.what() << endl;
		return 1;
	}
	const auto fuzz_config_yaml = _fuzz_config_yaml; // do not use _fuzz_config_yaml anymore
	


	// 
	// sanity check and load app config
	//
	if(app_config_filename == "" )
	{
		cout<< "Must specify an application config file " << endl;
		Usage(argc,argv, upper_opts);
		return 1;

	}
	auto _app_config_yaml=YAML::Node();
	try
	{
		_app_config_yaml=YAML::LoadFile(app_config_filename);
	}
	catch(const exception& e)
	{
		cout << "Could not parse " << app_config_filename << endl;
		cout << e.what() << endl;
		return 1;
	}
	const auto app_config_yaml=_app_config_yaml; // do not use _app_config_yaml anymore


	// 
	// sanity check version number and download zafld binary
	//
	if(ver_id < 0)
	{
		cout<< "Invalid or unspecified version id " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	// 
	// how to download a zafld file.
	//
	const auto download_file=[&](const string& filename, const uint32_t ver_id) -> int
		{
			SETUP_CONN(GetZafldRequest_t,ZafldBinary_t);

			request.set_ver_id(ver_id);

			while(true)
			{
				const auto status = backend->GetZafld(&context, request, &reply);

				// Act upon its status.
				if (!status.ok()) 
				{
					print_error(status);
					return 1;
				}
				if(reply.err())
				{
					cout << "Detected error in creating fuzz-ready version" << endl;
					return 1;
				}
				if(reply.zafld_bin() != "" )
					break;
				cout << "Binary not ready, sleeping 5 seconds..." << endl;
				sleep(5);
			}
			cout << "Server returned OK on getting Zafl'd binary with size " << reply.zafld_bin().size() << endl;

			ofstream strm(filename, ios::binary);
			strm << reply.zafld_bin();
			strm.close();
			return 0;
		};



	// 
	// download the zafld file and check for err.
	//
	const auto df_res = download_file("/tmp/to_fuzz.exe", ver_id);
	if(df_res !=0 )
	{
		return df_res;
	}


	// 
	// log some stuff before we start actually fuzzing
	//
	cout << "The fuzz config is: " << fuzz_config_yaml << endl;
	cout << "The app config is: "  << app_config_yaml  << endl;


	// start afl-fuzz

	return 1;
}

int LogDelete(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

int LogGetPrep(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, log get prep requires version id " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>2)
	{
		cout << "Error, junk at end of log get prep command " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	SETUP_CONN(GetZaflLogRequest_t,LogFile_t);

	request.set_ver_id(atoi(argv[1]));

	const auto status = backend->GetZaflLog(&context, request, &reply);

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}
	if(reply.log() == "" )
	{
		cout << "Log is currently empty." << endl;
	}
	else
	{
		cout << "The current prep. log is: " << endl << reply.log() << endl;
	}
	return 1;
}

int LogGetFuzz(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

int LogGet(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"prep", LogGetPrep },
			{"fuzz", LogGetFuzz },
		};

	return subcommandDispatch("log get", subcommands, argc, argv, upper_opts);
}

//
// process the boost subcommand
//
int Log(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"get",    LogGet    },
			{"delete", LogDelete },
		};

	return subcommandDispatch("log", subcommands, argc, argv, upper_opts);
}

// 
// handle processing of the turbo cli
//
int main(int argc, char *argv[])
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"ping",    Ping    },
			{"boost",   Boost   },
			{"version", Version },
			{"log",    Log     },
			{"fuzz",    Fuzz    },
			{"--help",  Usage   },
			{"help",    Usage   },
			{"--usage", Usage   },
			{"usage",   Usage   },
			{"-h",      Usage   },
			{"-?",      Usage   },
		};
	return subcommandDispatch("turbo-cli", subcommands, argc, argv, {{"pgm-name", argv[0]}, {"host", "localhost:55155"}});
}
